from typing import Literal

TemperatureType = Literal["F", "C"]
WindSpeedType = Literal["M", "K"]


def f_to_c(f: float) -> float:
    """
    Converts Fahrenheit to Celcius.

    Parameters
    ----------
    f : float
        Temperature in Fahrenheit

    Returns
    -------
    float
        Temperature in Celcius
    """
    return (f - 32) * (5/9)


def c_to_f(c: float) -> float:
    """
    Converts Celcius to Fahrenheit.

    Parameters
    ----------
    c : float
        _description_

    Returns
    -------
    float
        _description_
    """
    return (c * (5/9)) + 32


def process_temp(temp: float, format: TemperatureType) -> float:
    """
    Standardized temperature. Converts `temp` to Celcius if `format` is `F`
    and returns `temp` if `format` is `C`.

    Parameters
    ----------
    temp : float
        Temperature
    format : TemperatureType
        Boltwood II Format

    Returns
    -------
    float
        Temperature in Celcius
    """
    if (format == "C"):
        return temp
    else:
        return f_to_c(temp)


def knots_to_mph(k: float) -> float:
    """
    Converts knots to MPH.

    Parameters
    ----------
    k : float
        Speed in knots

    Returns
    -------
    float
        Speed in MPH
    """
    return k * 1.151


def mph_to_knots(k: float) -> float:
    """
    Converts MPH to knots.

    Parameters
    ----------
    k : float
        Speed in MPH

    Returns
    -------
    float
        Speed in knots
    """
    return k / 1.151


def process_speed(speed: float, format: WindSpeedType) -> float:
    """
    Standardized speed. Converts `speed` to MPH if `format` is `K`
    and returns `speed` if format is `M`.

    Parameters
    ----------
    speed : float
        Speed
    format : WindSpeedType
        Boltwood II Format

    Returns
    -------
    float
        Speed in MPH
    """
    if (format == "M"):
        return speed
    else:
        return knots_to_mph(speed)


class WeatherParseError(Exception):
    pass



from enum import Enum


class CloudyFlags(Enum):
    clear = 1
    light = 2
    very_cloudy = 3


class WindLimitFlags(Enum):
    calm = 1
    windy = 2
    very_windy = 3


class RainFlags(Enum):
    dry = 1
    damp = 2
    rain = 3


class DarknessFlags(Enum):
    dark = 1
    dim = 2
    daylight = 3



from datetime import datetime
from typing import cast

from .flags import *
from .utils import *


class WeatherEntry:
    """
    A class for parsing and storing a single weather entry.
    """

    time: datetime
    """Time at which this entry occured"""
    
    temp_scale: TemperatureType
    """Temperature format"""
    wind_scale: WindSpeedType
    """Wind speed format"""

    sky_temperature: float
    """Sky temperature in specified format"""
    ambient_temperature: float
    """Ambient temperature in specified format"""
    sensor_temperature: float
    """Sensor temperature in specified format"""
    wind_speed: float
    """Wind speed in specified format"""
    humidity: float
    """Humidity"""
    dew_point: float
    """Dewpoint in specified format"""
    dew_heater_percentage: float
    """Dew heater percentage"""

    is_raining: bool
    """Is raining"""
    is_wet: bool
    """Is wet"""

    cloudy: CloudyFlags
    """Cloud status"""
    wind_limit: WindLimitFlags
    """Wind status"""
    rain: RainFlags
    """Rain status"""
    darkness: DarknessFlags
    """Darkness status"""

    roof_closed: bool
    """Is the roof closed"""
    alert: bool
    """Is there an alert"""

    def __init__(self, entry: str):
        """
        Initialize object with a weather entry.

        Parameters
        ----------
        entry : str
            Raw weather entry

        Raises
        ------
        WeatherParseError
            Invalid format
        """

        parsed = entry.split(" ")
        while "" in parsed:
            parsed.remove("")

        self.time = datetime.strptime(
            parsed[0] + " " + parsed[1], "%Y-%m-%d %H:%M:%S.%f")

        self.temp_scale = cast(TemperatureType, parsed[2])
        self.wind_scale = cast(WindSpeedType, parsed[3])

        if self.temp_scale not in ["F", "C"]:
            raise WeatherParseError(f"Invalid temperature scale: {self.temp_scale}")

        if self.wind_scale not in ["M", "K"]:
            raise WeatherParseError(f"Invalid wind speed scale: {self.wind_scale}")

        self.sky_temperature = float(parsed[4])
        self.ambient_temperature = float(parsed[5])
        self.sensor_temperature = float(parsed[6])
        self.wind_speed = float(parsed[7])
        self.humidity = float(parsed[8])
        self.dew_point = float(parsed[9])
        self.dew_heater_percentage = float(parsed[10])

        self.is_raining = int(parsed[11]) == 1
        self.is_wet = int(parsed[12]) == 1

        self.cloudy = CloudyFlags(int(parsed[15]))
        self.wind_limit = WindLimitFlags(int(parsed[16]))
        self.rain = RainFlags(int(parsed[17]))
        self.darkness = DarknessFlags(int(parsed[18]))
        self.roof_closed = int(parsed[19]) == 1
        self.alert = int(parsed[20]) == 1



from datetime import datetime
import math
from .entry import WeatherEntry


class EntryCollection:
    """
    A class for parsing, storing, and processing multiple weather entries.
    """

    entries: list[WeatherEntry] = []
    """Entries"""

    def __init__(self, data: str, sort: bool = False) -> None:
        """
        Start parsing through weather entries that are separated by new lines.
        If `sort` is `False`, then the data will not be sorted. In order for
        other processing methods to work, the data must already be in order.

        Parameters
        ----------
        data : str
            Raw weather data. Could be read from a file.
        sort : bool, optional
            Sort the data by time, by default `False`
        """
        lines = data.splitlines()
        for i in lines:
            if i.strip() == "":
                continue
            self.entries.append(self.process_entry(i))

        if sort:
            self.entries.sort(key=lambda x: x.time.timestamp())

    def process_entry(self, line: str) -> WeatherEntry:
        """
        Process and entry before storing it. Override this to use custom
        weather entry objects.

        Parameters
        ----------
        line : str
            Raw entry

        Returns
        -------
        WeatherEntry
            Processed entry
        """
        return WeatherEntry(line)
    
    def find(self, time: datetime) -> WeatherEntry:
        """
        Finds the closest weather entry to this time.

        Parameters
        ----------
        time : datetime
            Time

        Returns
        -------
        WeatherEntry
            Weather entry closest to `time`
        """
        if time < self.entries[0].time:
            return self.entries[0]
        elif time > self.entries[-1].time:
            return self.entries[-1]
        
        return self.__find(time)
        
        
    def __find(self, time: datetime, _low: int = 0, _high: int = -1) -> WeatherEntry:
        """
        Internal method for searching. Does not account for first or last entry.
        """
        if _high == -1:
            _high = len(self.entries) - 1
            
        if _high == _low:
            found = self.entries[_high]
            next_to_found = self.entries[_high - 1]
            if abs(time.timestamp() - found.time.timestamp()) < abs(time.timestamp() - next_to_found.time.timestamp()):
                return found
            else:
                return next_to_found
        
        mid = math.floor(_low + ((_high - _low) / 2))
        if self.entries[mid].time > time:
            return self.__find(time, _low, mid)
        else:
            return self.__find(time, mid + 1, _high)
